#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from IPython.core.pylabtools import print_figure
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
import itertools
import codecs
import numpy as np


from segment import Segment

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Notebook:
	pass

_notebook = Notebook()
_notebook.extent = None
_DEFAULT_NOTEBOOK_WIDTH = 10
_notebook.width = _DEFAULT_NOTEBOOK_WIDTH

def set_notebook_crop(segment=None, margin=0.1):

	if segment is None:
		_notebook.extent = None
	else:
		assert isinstance(segment, Segment)
		assert segment
		duration = segment.duration
		_notebook.extent = Segment(
			segment.start-margin*duration, 
			segment.end+margin*duration
		)

def set_notebook_width(inches=None):
	if inches is None:
		_notebook.width = _DEFAULT_NOTEBOOK_WIDTH
	else:
		_notebook.width = inches

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _setup():
	fig, ax = plt.subplots()
	ax.set_xlim(_notebook.extent)
	ax.set_xlabel('Time')
	ax.set_ylim((0, 1))
	ax.axes.get_yaxis().set_visible(False)
	return fig, ax

def _render(fig):
	data = print_figure(fig, 'png')
	plt.close(fig)
	return data

def _draw_segment(ax, segment, y, color, label=None):
	if segment:
		ax.hlines(y, segment.start, segment.end, color, lw=1)
		ax.vlines(segment.start, y+0.05, y-0.05, color, lw=1)
		ax.vlines(segment.end, y+0.05, y-0.05, color, lw=1)
		if label:
			text = ax.text(segment.middle, y+0.05, codecs.encode(unicode(label), 'ascii', 'replace'),
				horizontalalignment='center', fontsize=10)
			# text.draw()
			# [[x0, y0], [x1, y1]] = text.get_window_extent().get_points()

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def repr_segment(segment):
	
	# remember current figure size 
	figsize = plt.rcParams['figure.figsize']
	# and update it for segment display
	plt.rcParams['figure.figsize'] = (_notebook.width, 1)

	if not _notebook.extent:
		set_notebook_crop(segment=segment)

	fig, ax = _setup()
	
	y = 0.5
	color = 'b'
	_draw_segment(ax, segment, y, color)

	data = _render(fig)

	# go back to previous figure size
	plt.rcParams['figure.figsize'] = figsize

	return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _y(segments):
	"""

	Parameters
	----------
	segments : iterator
		`Segment` iterator (sorted)

	Returns
	-------
	y : np.array
		y coordinates of each segment

	"""

	# up_to stores the largest end time 
	# displayed in each line (at the current iteration)
	# (at the beginning, there is only one empty line)
	up_to = [-np.inf]

	# y[k] indicates on which line to display kth segment
	y = []

	for segment in segments:
		# so far, we do not know which line to use
		found = False
		# try each line until we find one that is ok
		for i, u in enumerate(up_to): 
			# if segment starts after the previous one
			# on the same line, then we add it to the line
			if segment.start >= u:
				found = True
				y.append(i)
				up_to[i] = segment.end
				break
		# in case we went out of lines, create a new one
		if not found:
			y.append(len(up_to))
			up_to.append(segment.end)

	# from line numbers to actual y coordinates
	y = 1. - 1./(len(up_to)+1) * (1+np.array(y))

	return y

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def repr_timeline(timeline):

	# remember current figure size 
	figsize = plt.rcParams['figure.figsize']
	# and update it for segment display
	plt.rcParams['figure.figsize'] = (_notebook.width, 1)

	if not _notebook.extent and timeline:
		set_notebook_crop(segment=timeline.extent())

	cropped = timeline.crop(_notebook.extent, mode='loose')

	fig, ax = _setup()
	
	color = 'b'
	for segment, y in itertools.izip(cropped, _y(cropped)):
		_draw_segment(ax, segment, y, color)

	data = _render(fig)

	# go back to previous figure size
	plt.rcParams['figure.figsize'] = figsize

	return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def repr_annotation(annotation):

	# remember current figure size 
	figsize = plt.rcParams['figure.figsize']
	# and update it for segment display
	plt.rcParams['figure.figsize'] = (_notebook.width, 2)

	if not _notebook.extent:
		set_notebook_crop(segment=annotation.get_timeline().extent())

	cropped = annotation.crop(_notebook.extent, mode='loose')
	segments = [s for s, _ in cropped.itertracks()]

	# one color per label
	chart = cropped.chart()
	cm = get_cmap('gist_rainbow')
	colors = {
		label: cm(1.*i/len(chart))
		for i, (label, _) in enumerate(chart) 
	}

	fig, ax = _setup()

	for (segment, track, label), y in itertools.izip(
		cropped.itertracks(label=True), _y(segments)):
		color = colors[label]  # color = f(label)
		_draw_segment(ax, segment, y, color, label=label)

	data = _render(fig)

	# go back to previous figure size
	plt.rcParams['figure.figsize'] = figsize

	return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~




#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class NoMatch(object):
    """
    Parameters
    ----------
    format : str, optional
            
    """
    
    nextID = 0
    """
    Keep track of the number of instances since last reset
    """
    
    @classmethod
    def reset(cls):
        """Reset counter"""
        cls.nextID = 0
    
    @classmethod
    def next(cls):
        """Increment & get counter"""
        cls.nextID += 1
        return cls.nextID
    
    def __init__(self, format='NoMatch%03d'):
        super(NoMatch, self).__init__()
        self.ID = NoMatch.next()
        self.format = format
    
    def __str__(self):
        return self.format % self.ID
    
    def __repr__(self):
        return 'Ø'
        
    def __hash__(self):
        return hash(self.ID)
        
    def __eq__(self, other):
        if isinstance(other, NoMatch):
            return self.ID == other.ID
        else:
            return False

class MElement(object):
    
    def __init__(self, modality, element):
        super(MElement, self).__init__()
        self.modality = modality
        self.element = element
    
    def __eq__(self, other):
        return (self.element == other.element) & \
               (self.modality == other.modality)
    
    def __hash__(self):
        return hash(self.element)
    
    def __str__(self):
        return '%s (%s)' % (self.element, self.modality)
    
    def __repr__(self):
        return str(self)

class Mapping(object):
    """Many-to-many label mapping
    
    Parameters
    ----------
    left : str, optional
        Left-hand side modality. Defaults to 'left'.
    right : str
        Right-hand side modality. Defaults to 'right'.
    
    Returns
    -------
    mapping : Mapping
    
    Examples
    --------
    
        >>> mapping = Mapping('speaker', 'face')
        >>> mapping += (('A', 'B', 'C'), ('a', 'b'))
        >>> mapping += (('D', ), ('c', 'd', e))
        >>> print mapping
        (
           A B C --> a b
           D --> c d e
        )
    
    """
    def __init__(self, left="left", right="right"):
        super(Mapping, self).__init__()
        
        # left & right modality
        self.__left = left
        self.__right = right
        
        # left-to-right one-to-many label mapping
        # { left_label --> sorted_right_labels_tuple }
        # 
        self._left_to_right = {}
        
        # right-to-left one-to-many label mapping
        # { right_label --> sorted_left_labels_tuple }
        self._right_to_left = {}
        
        # left-to-right many-to-many label mapping
        # { sorted_left_labels_tuple --> sorted_right_labels_tuple }
        self._many_to_many = {}
    
    def __get_left(self): 
        return self.__left
    left = property(fget=__get_left)
    """Left-hand side modality"""
    
    def __get_right(self): 
        return self.__right
    right = property(fget=__get_right)
    """Right-hand side modality"""
    
    def __get_left_set(self):
        return set(self._left_to_right)
    left_set = property(fget=__get_left_set)
    """Left-hand side set of labels"""
    
    def __get_right_set(self):
        return set(self._right_to_left)
    right_set = property(fget=__get_right_set)
    """Right-hand side set of labels"""
    
    # Normalize left_right argument to 
    # ( tuple(sorted_left_labels), tuple(sorted_right_labels) )
    def _check_and_normalize(self, left_right):
        
        if not isinstance(left_right, (tuple, list)) or len(left_right) != 2:
            raise ValueError('expected 2-tuple of list of labels')
        
        left = left_right[0]
        right = left_right[1]
        
        if left is None:
            left = tuple()
        elif isinstance(left, (list, tuple, set)):
            left = tuple(sorted(left))
        else:
            raise ValueError('expected list, tuple or set for left part.')
        
        if right is None:
            right = tuple()
        elif isinstance(right, (list, tuple, set)):
            right = tuple(sorted(right))
        else:
            raise ValueError('expected list, tuple or set for right part.')
        
        return left, right
    
    def empty(self):
        """Empty copy"""
        M = type(self)(self.left, self.right)
        return M
    
    def copy(self):
        """Duplicate mapping"""
        
        # starts with an empty mapping
        mapping = self.empty()
        
        # add 
        for left, right in self:
            mapping += (left, right)
        return mapping
    
    def __add__(self, other):
        """
        
        Use expression 'mapping + (left_labels, right_labels)'
                    or 'mapping + other_mapping'
        
        Parameters
        ----------
        other : Mapping or couple of labels
        
        Returns
        -------
        mapping : Mapping
        
        Examples
        --------
        
            >>> mapping1 = Mapping()
            >>> mapping1 += (('A', 'B', 'C'), ('a', 'b'))
            >>> mapping2 = Mapping()
            >>> mapping2 += (('D',), ('c', 'd', 'e'))
            >>> mapping = mapping1 + mapping2
            >>> print mapping
            (
               A B C --> a b
               D --> c d e
            )
        
        """
        
        mapping = self.copy()
        mapping += other
        return mapping
    
    def __iadd__(self, other):
        """
        
        Use expression 'mapping += (left_labels, right_labels)'
                    or 'mapping += other_mapping'
                    
        Examples
        --------

            >>> mapping = Mapping()
            >>> mapping += (('A', 'B', 'C'), ('a', 'b'))
            >>> print mapping
            (
               A B C --> a b
            )
            >>> other_mapping = Mapping()
            >>> other_mapping += (('D',), ('c', 'd', 'e'))
            >>> mapping += other_mapping
            >>> print mapping
            (
               A B C --> a b
               D --> c d e
            )
            
        Raises
        ------
        ValueError
        
        """
        
        if isinstance(other, Mapping):
            if other.left != self.left or other.right != self.right:
               raise ValueError('incompatible mapping modalities')
            
            for left, right in other:
                self += (left, right)
            return self
        
        left, right = self._check_and_normalize(other)
        
        already_mapped = set(left) & self.left_set
        if already_mapped:
            already_mapped = already_mapped.pop()
            raise ValueError('%s (%s) is already mapped to %s.' % \
                             (already_mapped, self.left, \
                              self._left_to_right[already_mapped]))
            
        already_mapped = set(right) & self.right_set
        if already_mapped:
            already_mapped = already_mapped.pop()
            raise ValueError('%s (%s) is already mapped to %s.' % \
                             (already_mapped, self.right, \
                              self._right_to_left[already_mapped]))
        
        for label in left:
            self._left_to_right[label] = right
        for label in right:
            self._right_to_left[label] = left
        
        if left == tuple():
            left = NoMatch()
        if right == tuple():
            right = NoMatch()
        self._many_to_many[left] = right
        
        return self
        
    # --- iterator ---
    
    def __iter__(self):
        """Left/right mapping iterator
        
        Examples
        --------
        
            >>> mapping = Mapping()
            >>> mapping += (('A', 'B', 'C'), ('a', 'b'))
            >>> mapping += (('D',), ('c', 'd', 'e'))
            >>> for left, right in mapping:
            ...    print '%s --> %s' % (' '.join(sorted(left)), \
                                        ' '.join(sorted(right)))
            A B C --> a b
            D --> c d e
            
        """
        
        for left, right in self._many_to_many.iteritems():
            
            if isinstance(left, NoMatch):
                left = set([])
            else:
                left = set(left)
            
            if isinstance(right, NoMatch):
                right = set([])
            else:
                right = set(right)
            
            yield left, right
    
    def __contains__(self, label):
        """Use expression 'label in mapping'"""
        return label in self.left_set
    
    def __hash__(self):
        """Use expression hash(mapping)"""
        return hash(tuple(sorted(self.left_set))) + \
               hash(tuple(sorted(self.right_set)))
    
    def __eq__(self, other):
        """Use expression 'mapping == other_mapping'"""
        return self._left_to_right == other._left_to_right and \
               self._right_to_left == other._right_to_left
        
    def to_partition(self):
        
        partition = {}
        C = 0
        for left, right in self:
            partition.update({MElement(self.left, element): C \
                              for element in left})
            partition.update({MElement(self.right, element): C \
                              for element in right})
            C += 1
        return partition
        
    def to_expected_partition(self):
        
        left = self.left_set
        right = self.right_set
        expected = {element:e for e, element in enumerate(left | right)}

        partition = {}
        for element in left:
            partition[MElement(self.left, element)] = expected[element]
        for element in right:
            partition[MElement(self.right, element)] = expected[element]
        
        return partition
        
    def to_dict(self, reverse=False):
        if reverse:
            return {(tuple(left) if left else NoMatch()): \
                    (tuple(right) if right else NoMatch()) \
                    for right, left in self}
        else:
            return {(tuple(left) if left else NoMatch()): \
                    (tuple(right) if right else NoMatch()) \
                    for left, right in self}
    
    def to_expected_dict(self, reverse=False):
        
        left = self.left_set
        right = self.right_set
        both = left & right
        
        expected_dict = {}
        for element in both:
            expected_dict[element] = element
        for element in left-both:
            expected_dict[element] = NoMatch()
        for element in right-both:
            expected_dict[NoMatch()] = element
        
        if reverse:
            return {value:key for key, value in expected_dict.iteritems()}
        else:
            return expected_dict
    
    def __getitem__(self, key):
        return self._left_to_right[key]
    
    def __str__(self):
        """Human-readable representation
        
        Examples
        --------
        
            >>> mapping = Mapping()
            >>> mapping += (('A', 'B', 'C'), ('a', 'b'))
            >>> mapping += (('D',), ('c', 'd', 'e'))
            >>> print mapping
            (
               A B C --> a b
               D --> c d e
            )
        """
        
        string = "(\n"
        for left, right in self:
            left = [str(i) for i in left]
            right = [str(i) for i in right]
            string += '   %s <--> %s\n' % (' '.join(sorted(left)), \
                                           ' '.join(sorted(right)))
        string += ")"
        return string

class ManyToOneMapping(Mapping):
    """Many-to-one mapping"""
    
    def _check_and_normalize(self, mapping):
        """
        Extra verification for many-to-one mapping
        """
        elements1, elements2 = super(ManyToOneMapping, self)\
                               ._check_and_normalize(mapping)
        
        if len(elements2) > 1:
            raise ValueError('Right mapping part (%s) must contain only one element.' % elements2)
        
        return elements1, elements2
    
    @classmethod
    def fromMapping(cls, mapping):
        M = cls(mapping.left, mapping.right)
        for l, r in mapping:
            M += (l, r)
        return M
        
    def __getitem__(self, key):
        right = self._left_to_right[key]
        if right:
            return right[0]
        else:
            return None
            
    def __call__(self, key):
        return self[key]      

class OneToOneMapping(ManyToOneMapping):
    """One-to-one mapping"""
    
    def _check_and_normalize(self, mapping):
        """
        Extra verification for one-to-one mapping
        """
        elements1, elements2 = super(OneToOneMapping, self)._check_and_normalize(mapping)
        
        if len(elements1) > 1:
            raise ValueError('Left mapping part (%s) must contain only one element.' % elements1)
            
        return elements1, elements2
    
    @classmethod
    def fromMapping(cls, mapping):
        M = cls(mapping.left, mapping.right)
        for l, r in mapping:
            M += (l, r)
        return M
    
    def to_dict(self, reverse=False, single=False):
        D = super(OneToOneMapping, self).to_dict(reverse=reverse)
        if single:
            d = {left if isinstance(left, NoMatch) else left[0]: \
                 right if isinstance(right, NoMatch) else right[0] \
                 for left, right in D.iteritems()}
            return d
        else:
            return D


if __name__ == "__main__":
    import doctest
    doctest.testmod()




#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from pandas import MultiIndex, DataFrame, pivot_table

from . import PYANNOTE_SEGMENT, PYANNOTE_TRACK, PYANNOTE_LABEL, PYANNOTE_SCORE
from annotation import Annotation, Unknown
from segment import Segment
from timeline import Timeline


class AnnotationMixin(object):

    def get_timeline(self):
        if self._timelineHasChanged:
            self._timeline = Timeline(segments=[s for s, _ in self._df.index],
                                      uri=self.uri)
            self._timelineHasChanged = False
        return self._timeline

    def __len__(self):
        """Number of annotated segments"""
        return len(self.get_timeline())

    def __nonzero__(self):
        """False if annotation is empty"""
        return len(self.get_timeline()) > 0

    def __contains__(self, included):
        """Check if segments are annotated

        Parameters
        ----------
        included : `Segment` or `Timeline`

        Returns
        -------
        contains : bool
            True if every segment in `included` is annotated, False otherwise.
        """
        return included in self.get_timeline()

    def __iter__(self):
        """Iterate over sorted segments"""
        return iter(self.get_timeline())

    def __reversed__(self):
        """Reverse iterate over sorted segments"""
        return reversed(self.get_timeline())

    def itersegments(self):
        return iter(self.get_timeline())

    def itertracks(self):
        """Iterate over annotation as (segment, track) tuple"""

        # make sure segment/track pairs are sorted
        self._df = self._df.sort_index()

        for (segment, track), _ in self._df.iterrows():
            yield segment, track

    def crop(self, focus, mode='strict'):
        """Crop on focus

        Parameters
        ----------
        focus : `Segment` or `Timeline`

        mode : {'strict', 'loose', 'intersection'}
            In 'strict' mode, only segments fully included in focus coverage are
            kept. In 'loose' mode, any intersecting segment is kept unchanged.
            In 'intersection' mode, only intersecting segments are kept and
            replaced by their actual intersection with the focus.

        Returns
        -------
        cropped : same type as caller
            Cropped version of the caller containing only tracks matching
            the provided focus and mode.

        Remarks
        -------
        In 'intersection' mode, the best is done to keep the track names
        unchanged. However, in some cases where two original segments are
        cropped into the same resulting segments, conflicting track names are
        modified to make sure no track is lost.

        """
        if isinstance(focus, Segment):

            return self.crop(Timeline([focus], uri=self.uri), mode=mode)

        elif isinstance(focus, Timeline):

            # timeline made of all annotated segments
            timeline = self.get_timeline()

            # focus coverage
            # coverage = focus.coverage()
            coverage = focus.segmentation()

            if mode in ['strict', 'loose']:

                # segments (strictly or loosely) included in requested coverage
                included = timeline.crop(coverage, mode=mode)

                # boolean array: True if row must be kept, False otherwise
                keep = [(s in included) for s, _ in self._df.index]

                # crop-crop
                A = self.__class__(uri=self.uri, modality=self.modality)
                A._df = self._df[keep]

                return A

            elif mode == 'intersection':

                # two original segments might be cropped into the same resulting
                # segment -- therefore, we keep track of the mapping
                intersection, mapping = timeline.crop(coverage,
                                                      mode=mode, mapping=True)

                # create new empty annotation
                A = self.__class__(uri=self.uri, modality=self.modality)

                for cropped in intersection:
                    for original in mapping[cropped]:
                        for track in self.tracks(original):
                            # try to use original track name (candidate)
                            # if it already exists, create a brand new one
                            new_track = A.new_track(cropped, candidate=track)
                            # copy each value, column by column
                            for label in self._df.columns:
                                value = self._df.get_value((original, track),
                                                           label)
                                A._df = A._df.set_value((cropped, new_track),
                                                        label, value)

                return A

        else:
            raise TypeError('')

    def tracks(self, segment):
        """Set of tracks for query segment

        Parameters
        ----------
        segment : `Segment`
            Query segment

        Returns
        -------
        tracks : set
            Set of tracks for query segment
        """

        try:
            df = self._df.xs(segment)
            existing_tracks = set(df.index)

        except Exception, e:
            existing_tracks = set([])

        return existing_tracks

    def has_track(self, segment, track):
        """Check whether a given track exists

        Parameters
        ----------
        segment : `Segment`
            Query segment
        track :
            Query track

        Returns
        -------
        exists : bool
            True if track exists for segment
        """
        return (segment, track) in self._df.index

    def get_track_by_name(self, track):
        """Get all tracks with given name

        Parameters
        ----------
        track : any valid track name
            Requested name track

        Returns
        -------
        tracks : list
            List of (segment, track) tuples
        """
        try:
            segments = list(self._df.xs(track, level=1).index)
        except Exception, e:
            segments = []
        return [(s, track) for s in segments]

    def copy(self):
        A = self.__class__(uri=self.uri, modality=self.modality)
        A._df = self._df.copy()
        return A

    def retrack(self):
        """
        """
        A = self.copy()
        reindex = MultiIndex.from_tuples([(s, n)
                                          for n, (s, _) in enumerate(A._df.index)])
        A._df.index = reindex
        return A

    def new_track(self, segment, candidate=None, prefix=None):
        """Track name generator

        Parameters
        ----------
        segment : Segment
        prefix : str, optional
        candidate : any valid track name


        Returns
        -------
        track : str
            New track name
        """

        # obtain list of existing tracks for segment
        existing_tracks = self.tracks(segment)

        # if candidate is provided, check whether it already exists
        # in case it does not, use it
        if candidate is not None:
            if candidate not in existing_tracks:
                return candidate

        # no candidate was provided or the provided candidate already exists
        # we need to create a brand new one

        # by default (if prefix is not provided)
        # use modality as prefix (eg. speaker1, speaker2, ...)
        if prefix is None:
            prefix = '' if self.modality is None else str(self.modality)

        # find first non-existing track name for segment
        # eg. if speaker1 exists, try speaker2, then speaker3, ...
        count = 1
        while ('%s%d' % (prefix, count)) in existing_tracks:
            count += 1

        # return first non-existing track name
        return '%s%d' % (prefix, count)

    def __str__(self):
        """Human-friendly representation"""
        if self:
            self._df.sort_index(inplace=True)
            return str(self._df)
        else:
            return ""


class Scores(AnnotationMixin, object):
    """

    Parameters
    ----------
    uri : str, optional

    modality : str, optional

    Returns
    -------
    scores : `Scores`

    Examples
    --------

        >>> s = Scores(uri='video', modality='speaker')
        >>> s[Segment(0,1), 's1', 'A'] = 0.1
        >>> s[Segment(0,1), 's1', 'B'] = 0.2
        >>> s[Segment(0,1), 's1', 'C'] = 0.3
        >>> s[Segment(0,1), 's2', 'A'] = 0.4
        >>> s[Segment(0,1), 's2', 'B'] = 0.3
        >>> s[Segment(0,1), 's2', 'C'] = 0.2
        >>> s[Segment(2,3), 's1', 'A'] = 0.2
        >>> s[Segment(2,3), 's1', 'B'] = 0.1
        >>> s[Segment(2,3), 's1', 'C'] = 0.3

    """
    @classmethod
    def from_df(
        cls, df,
        uri=None, modality=None, aggfunc=np.mean
    ):
        """

        Parameters
        ----------
        df : DataFrame
            Must contain the following columns:
            'segment', 'track', 'label' and 'value'
        uri : str, optional
            Resource identifier
        modality : str, optional
            Modality
        aggfunc : func
            Value aggregation function in case of duplicate (segment, track,
            label) tuples

        Returns
        -------

        """
        A = cls(uri=uri, modality=modality)
        A._df = pivot_table(
            df, values=PYANNOTE_SCORE,
            rows=[PYANNOTE_SEGMENT, PYANNOTE_TRACK], cols=PYANNOTE_LABEL,
            aggfunc=aggfunc
        )
        return A

    def __init__(self, uri=None, modality=None):
        super(Scores, self).__init__()

        index = MultiIndex(
            levels=[[], []], labels=[[], []],
            names=[PYANNOTE_SEGMENT, PYANNOTE_TRACK]
        )

        self._df = DataFrame(index=index, dtype=np.float64)
        self.modality = modality
        self.uri = uri
        self._timelineHasChanged = True

    # del scores[segment]
    # del scores[segment, :]
    # del scores[segment, track]
    def __delitem__(self, key):

        if isinstance(key, Segment):
            segment = key
            self._df = self._df.drop(segment, axis=0)
            self._timelineHasChanged = True

        elif isinstance(key, tuple) and len(key) == 2:
            segment, track = key
            self._df = self._df.drop((segment, track), axis=0)
            self._timelineHasChanged = True

        else:
            raise KeyError('')

    # value = scores[segment, track, label]
    def __getitem__(self, key):
        segment, track, label = key
        return self._df.get_value((segment, track), label)

    def get_track_scores(self, segment, track):
        """Get all scores for a given track.

        Parameters
        ----------
        segment : Segment
        track : hashable
            segment, track must be a valid track

        Returns
        -------
        scores : dict
            {label: score} dictionary
        """
        return {l: self._df.get_value((segment, track), l) for l in self._df}

    # scores[segment, track, label] = value
    def __setitem__(self, key, value):
        segment, track, label = key
        self._df = self._df.set_value((segment, track), label, value)
        self._timelineHasChanged = True

    def labels(self, unknown=True):
        """List of labels

        Parameters
        ----------
        unknown : bool, optional
            When False, do not return Unknown instances
            When True, return any label (even Unknown instances)

        Returns
        -------
        labels : list
            Sorted list of existing labels

        Remarks
        -------
            Labels are sorted based on their string representation.
        """
        labels = sorted(self._df.columns, key=str)
        if unknown:
            return labels
        else:
            return [l for l in labels if not isinstance(l, Unknown)]

    def itervalues(self):
        """Iterate over annotation as (segment, track, label, value) tuple"""

        # make sure segment/track pairs are sorted
        self._df = self._df.sort_index()

        # yield one (segment, track, label) tuple per loop
        labels = self._df.columns
        for (segment, track), columns in self._df.iterrows():
            for label in labels:
                value = columns[label]
                if np.isnan(value):
                    continue
                else:
                    yield segment, track, label, value

    def _rank(self, invert):

        if invert:
            direction = 1.

        else:
            direction = -1.

        def nan_rank(data):

            # replace NaN by -inf or +inf depending on the requested direction
            finite = np.isfinite(data)
            fixed = np.where(finite, direction*data, -direction*np.inf)

            # do the actual argsort
            indices = np.argsort(fixed)
            # get rank from argsort
            rank = np.argsort(indices)

            # special treatment for inverted NaN scores
            # (we want ranks to start at 0 even in case of NaN)
            if invert:
                rank = np.where(finite, rank-(len(data)-np.sum(finite)), np.nan)
            else:
                rank = np.where(finite, rank, np.nan)
            return rank

        return self._df.apply(nan_rank, axis=1)

    def rank(self, invert=False):
        """

        Parameters
        ----------
        invert : bool, optional
            By default, larger scores are better.
            Set `invert` to True to indicate smaller scores are better.

        Returns
        -------
        rank : `Scores`

        """
        A = self.__class__(uri=self.uri, modality=self.modality)
        A._df = self._rank(invert)
        return A

    def nbest(self, n, invert=False):
        """

        Parameters
        ----------
        n : int
            Size of n-best list
        invert : bool, optional
            By default, larger scores are better.
            Set `invert` to True to indicate smaller scores are better.

        Returns
        -------
        nbest : `Scores`
            New scores where only n-best are kept.

        """
        df = self._df.copy()
        nbest = self._rank(invert) < n
        df[~nbest] = np.nan

        A = self.__class__(uri=self.uri, modality=self.modality)
        A._df = df

        return A

    def subset(self, labels, invert=False):
        """Scores subset

        Extract scores subset based on labels

        Parameters
        ----------
        labels : set
            Set of labels
        invert : bool, optional
            If invert is True, extract all but requested `labels`

        Returns
        -------
        subset : `Scores`
            Scores subset.
        """

        if not isinstance(labels, set):
            raise TypeError('labels must be provided as a set of labels.')

        if invert:
            labels = set(self.labels()) - labels
        else:
            labels = labels & set(self.labels())

        A = self.__class__(uri=self.uri, modality=self.modality)
        A._df = self._df[list(labels)]

        return A

    def to_annotation(self, threshold=-np.inf, posterior=False):
        """

        Parameters
        ----------
        threshold : float, optional
            Each track is annotated with the label with the highest score.
            Yet, if the latter is smaller than `threshold`, label is replaced
            with an `Unknown` instance.
        posterior : bool, optional
            If True, scores are posterior probabilities in open-set
            identification. If top model posterior is higher than unknown
            posterior, it is selected. Otherwise, label is replaced with an
            `Unknown` instance.
        """

        annotation = Annotation(uri=self.uri, modality=self.modality)
        if not self:
            return annotation

        best = self.nbest(1, invert=False)

        if posterior:

            # compute unknown posterior
            func = lambda p: 1. - np.nansum(p, axis=1)
            Pu = self.apply(func, new_columns=['_'])

            # threshold best target posterior
            # with unknown posterior and threshold
            for segment, track, label, value in best.itervalues():

                if value < Pu[segment, track, '_'] or value < threshold:
                    label = Unknown()

                annotation[segment, track] = label

        else:

            # threshold best target score with threshold
            for segment, track, label, value in best.itervalues():
                if value < threshold:
                    label = Unknown()
                annotation[segment, track] = label

        return annotation

    def map(self, func):
        """Apply function to all values"""
        A = self.__class__(uri=self.uri, modality=self.modality)
        A._df = func(self._df)
        return A

    def apply(self, data_func, new_index=None, new_columns=None):
        """Apply `data_func` on internal numpy array

        Parameters
        ----------
        data_func : func
            Function expecting (index x columns) numpy array as input
        new_index : iterable, optional
            When provided, these will be the index of returned array.
        new_columns : iterable, optional
            When provided, these will be the columns of returned array.
        """
        new_data = data_func(self._df.values)

        if new_index is None:
            new_index = self._df.index

        if new_columns is None:
            new_columns = self._df.columns

        df = DataFrame(
            data=new_data,
            index=new_index,
            columns=new_columns)

        new_scores = self.__class__(uri=self.uri, modality=self.modality)
        new_scores._df = df

        return new_scores


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""

Features.

"""

import numpy as np
from segment import Segment, SlidingWindow
from timeline import Timeline


class BaseSegmentFeature(object):

    """
    Base class for any segment/feature iterator.
    """

    def __init__(self, uri=None):
        super(BaseSegmentFeature, self).__init__()
        self.__uri = uri

    def __get_uri(self):
        return self.__uri

    def __set_uri(self, value):
        self.__uri = value
    uri = property(fget=__get_uri, fset=__set_uri)
    """Path to (or any identifier of) described resource"""

    def __iter__(self):
        """Segment/feature vector iterator

        Use expression 'for segment, feature_vector in segment_feature'

        This method must be implemented by subclass.

        """
        raise NotImplementedError('Missing method "__iter__".')


class BasePrecomputedSegmentFeature(BaseSegmentFeature):

    """'Segment iterator'-driven precomputed feature iterator.

    Parameters
    ----------
    data : numpy array
        Feature vectors stored in such a way that data[i] is ith feature vector.
    segment_iterator : :class:`SlidingWindow` or :class:`Timeline`
        Segment iterator.
        Its length must correspond to `data` length.
    uri : string, optional
        name of (audio or video) described resource

    """

    def __init__(self, data, segment_iterator, uri=None):
        # make sure data does not contain NaN nor inf
        data = np.asarray_chkfinite(data)

        # make sure segment_iterator is actually one of those
        if not isinstance(segment_iterator, (SlidingWindow, Timeline)):
            raise TypeError("segment_iterator must 'Timeline' or "
                            "'SlidingWindow'.")

        # make sure it iterates the correct number of segments
        try:
            N = len(segment_iterator)
        except Exception:
            # an exception is raised by `len(sliding_window)`
            # in case sliding window has infinite end.
            # this is acceptable, no worry...
            pass
        else:
            n = data.shape[0]
            if n != N:
                raise ValueError("mismatch between number of segments (%d) "
                                 "and number of feature vectors (%d)." % (N, n))

        super(BasePrecomputedSegmentFeature, self).__init__(uri=uri)
        self.__data = data
        self._segment_iterator = segment_iterator

    def __get_data(self):
        return self.__data
    data = property(fget=__get_data)
    """Raw feature data (numpy array)"""

    def __iter__(self):
        """Feature vector iterator

        Use expression 'for segment, feature_vector in periodic_feature'

        """

        # get number of feature vectors
        n = self.__data.shape[0]

        for i, segment in enumerate(self._segment_iterator):

            # make sure we do not iterate too far...
            if i >= n:
                break

            # yield current segment and corresponding feature vector
            yield segment, self.__data[i]

    def _segmentToRange(self, segment):
        """
        Parameters
        ----------
        segment : :class:`pyannote.base.segment.Segment`

        Returns
        -------
        i, n : int

        """
        raise NotImplementedError('Missing method "_segmentToRange".')

    def _rangeToSegment(self, i, n):
        """
        Parameters
        ----------
        i, n : int

        Returns
        -------
        segment : :class:`pyannote.base.segment.Segment`

        """
        raise NotImplementedError('Missing method "_rangeToSegment".')

    def __call__(self, subset, mode='loose'):
        """
        Use expression 'feature(subset)'

        Parameters
        ----------
        subset : :class:`pyannote.base.segment.Segment` or
                 :class:`pyannote.base.timeline.Timeline`

        Returns
        -------
        data : numpy array

        """

        if not isinstance(subset, (Segment, Timeline)):
            raise TypeError('')

        if isinstance(subset, Segment):
            i, n = self._segmentToRange(subset)
            indices = range(i, i + n)

        elif isinstance(subset, Timeline):
            indices = []
            for segment in subset.coverage():
                i, n = self._segmentToRange(segment)
                indices += range(i, i + n)

        return np.take(self.__data, indices, axis=0, out=None, mode='clip')


class PeriodicPrecomputedFeature(BasePrecomputedSegmentFeature):

    """'Sliding window'-driven precomputed feature iterator.

    Parameters
    ----------
    data : numpy array
        Feature vectors stored in such a way that data[i] is
        ith feature vector.
    sliding_window : :class:`SlidingWindow`
        Sliding window. Its length must correspond to `data` length
        (or it can be infinite -- ie. sliding_window.end = None)
    uri : string, optional
        name of (audio or video) described resource

    Examples
    --------
        >>> data = ...
        >>> sliding_window = SlidingWindow( ... )
        >>> feature_iterator = PeriodicPrecomputedFeature(data, sliding_window)
        >>> for segment, feature_vector in feature_iterator:
        ...     pass

    """

    def __init__(self, data, sliding_window, uri=None):

        super(PeriodicPrecomputedFeature, self).__init__(
            data, sliding_window, uri=uri
        )

    def __get_sliding_window(self):
        return self._segment_iterator
    sliding_window = property(fget=__get_sliding_window)

    def _segmentToRange(self, segment):
        """
        Parameters
        ----------
        segment : :class:`pyannote.base.segment.Segment`

        Returns
        -------
        i, n : int

        """
        return self.sliding_window.segmentToRange(segment)

    def _rangeToSegment(self, i, n):
        """
        Parameters
        ----------
        i, n : int

        Returns
        -------
        segment : :class:`pyannote.base.segment.Segment`

        """
        return self.sliding_window.rangeToSegment(i, n)


class TimelinePrecomputedFeature(BasePrecomputedSegmentFeature):

    """Timeline-driven precomputed feature iterator.

    Parameters
    ----------
    data : numpy array
        Feature vectors stored in such a way that data[i] is ith feature vector.
    timeline : :class:`Timeline`
        Timeline whose length must correspond to `data` length
    uri : string, optional
        name of (audio or video) described resource

    Examples
    --------
        >>> data = ...
        >>> timeline = Timeline( ... )
        >>> feature_iterator = TimelinePrecomputedFeature(data, timeline)
        >>> for segment, feature_vector in feature_iterator:
        ...     pass


    """

    def __init__(self, data, timeline, uri=None):
        super(TimelinePrecomputedFeature, self).__init__(data, timeline,
                                                         uri=uri)

    def __get_timeline(self):
        return self._segment_iterator
    timeline = property(fget=__get_timeline)

    def _segmentToRange(self, segment):
        timeline = self.timeline.crop(segment, mode='loose')
        if timeline:
            # index of first segment in sub-timeline
            first_segment = timeline[0]
            i = self.timeline.index(first_segment)
            # number of segments in sub-timeline
            n = len(timeline)
        else:
            i = 0
            n = 0

        return i, n

    def _rangeToSegment(self, i, n):
        first_segment = self.timeline[i]
        last_segment = self.timeline[i + n]
        return first_segment | last_segment


class SlidingWindowFeature(object):

    """Periodic feature vectors

    Parameters
    ----------
    data : (nSamples, nFeatures) numpy array

    sliding_window : SlidingWindow


    """

    def __init__(self, data, sliding_window):
        super(SlidingWindowFeature, self).__init__()
        self.sliding_window = sliding_window
        self.data = data

    def getNumber(self):
        """Number of feature vectors"""
        return self.data.shape[0]

    def getDimension(self):
        """Dimension of feature vectors"""
        return self.data.shape[1]

    def getExtent(self):
        return self.sliding_window.rangeToSegment(0, self.getNumber())

    def __getitem__(self, i):
        """Get ith feature vector"""
        return self.data[i]

    def iterfeatures(self, window=False):
        """Feature vector iterator

        Parameters
        ----------
        window : bool, optional
            When True, yield both feature vector and corresponding window.
            Default is to only yield feature vector

        """
        nSamples = self.data.shape[0]
        for i in xrange(nSamples):
            if window:
                yield self.data[i], self.sliding_window[i]
            else:
                yield self.data[i]

    def crop(self, focus):
        """Get set of feature vector for given segment

        Parameters
        ----------
        focus : Segment or Timeline

        Returns
        -------
        data : numpy array
            (nSamples, nFeatures) numpy array
        """

        n = self.getNumber()

        if isinstance(focus, Segment):
            firstFrame, frameNumber = self.sliding_window.segmentToRange(focus)
            indices = range(
                min(n, max(0, firstFrame)),
                min(n, max(0, firstFrame+frameNumber))
            )

        if isinstance(focus, Timeline):
            indices = []
            for segment in focus.coverage():
                firstFrame, frameNumber = self.sliding_window.segmentToRange(
                    segment)
                indices += range(
                    min(n, max(0, firstFrame)),
                    min(n, max(0, firstFrame+frameNumber))
                )

        return np.take(self.data, indices, axis=0, out=None, mode='clip')


if __name__ == "__main__":
    import doctest
    doctest.testmod()
